// democast.cpp  test cases for assert_cast template

#include <assert.h>
#include <iostream>
#include <typeinfo>
//#include "alt_cast.h"

using std::cout;
using std::endl;

// The following code was pasted in from cast.h,
// with some trace statements added to help
// show the templates in action.

#ifndef CAST_H_A2434B93
#define CAST_H_A2434B93

// for pointers
template<typename Result_ptr, typename Source>
inline Result_ptr assert_cast(Source* pSource)
{
        std::cout << "\npointer cast to " << typeid(Result_ptr).name()
                << "\nstatic source type: " << typeid(Source).name()
                << " *    dynamic source type: " << typeid(*pSource).name()
                << " *" << std::endl;
        assert ( static_cast<Result_ptr>(pSource) ==
                dynamic_cast<Result_ptr>(pSource) );
        return static_cast<Result_ptr>(pSource);
}

// for references
template<typename Result_ref, typename Source>
inline Result_ref assert_cast(Source& rSource)
{
        std::cout << "\nreference cast to " << typeid(Result_ref).name()
                << "\nstatic source type: " << typeid(Source).name()
                << " &    dynamic source type: " << typeid(rSource).name()
                << " &" << std::endl;
#ifndef NDEBUG
        try
        {
                assert ( &static_cast<Result_ref>(rSource) ==
                        &dynamic_cast<Result_ref>(rSource) );
        }
        catch(...)
        {
                // convert exceptions to assertions to prevent them from being "handled"
                assert(false);
        }
#endif
        return static_cast<Result_ref>(rSource);
}

#endif  // CAST_H_A2434B93

// [end code pasted from CAST.H]


class Base
{
public:
        virtual ~Base() {}
};

class Derived : public Base
{
};

Base* GetDerived()
{
        static Derived d;
        return &d;
}


int main()
{
        Derived d;
        const Base* pcd = &d;
        const Base& rcd = d;

        // legal casts, should not assert
        Derived* pd = assert_cast<Derived*>(GetDerived());
        const Derived* pcd2 = assert_cast<const Derived*>(pcd);
        Derived& rd = assert_cast<Derived&>(d);
        const Derived& rcd2 = assert_cast<const Derived&>(rcd);

        // each of these should cause a compile-time error if uncommented
//      const Derived& crd3 = assert_cast<Derived&>(crd);       // attempting to cast away const
//      Derived d2 = assert_cast<Derived>(d);   // not a ptr or reference

        Base b;
        Base& rb = b;
        Base* pb = &b;
        Base* const& crpb = pb;

        // each of these should raise an assertion if uncommented
        pd = assert_cast<Derived*>(crpb);
//      pd = assert_cast<Derived*>(&b);
//      Derived& rd = assert_cast<Derived&>(rb);

        (void)pd;
        (void)pcd2;
        (void)rd;
        (void)rcd2;
        (void)rb;
        cout << "Done." << endl;
        return 0 ;
}
